/*
 * WumpusMaze class captures the relevant aspects of the Wumpus world
 */

package edu.umn.d.cs5541.program3.wumpusworld;

/**
 * Class to represent a wumpus world maze
 */
public class WumpusMaze {
    // Represent contents of location in maze by setting bits
    private static final int GOLD_BIT = 1;
    private static final int PIT_BIT = 2;
    private static final int WUMPUS_BIT = 4;

    // Cost of some occurrences (based on the textbook)
    private static final int LOSE_COST = 1000;
    private static final int GOLD_REWARD = 1000;
    private static final int USE_ARROW_COST = 10;
    private static final int ACTION_COST = 1;

    // Important aspects of the maze
    private int maze[][] = null;
    private int mazeSize = 0;
    private boolean hasArrow = true;
    private int facingAngle = 0;
    private int xLocation = 0;
    private int yLocation = 0;
    private boolean gameOver = false;
    private int totalReward = 0;

    // Possible actions the user may execute
    public static final int FORWARD_ACTION = 1;
    public static final int LEFT_ACTION = 2;
    public static final int RIGHT_ACTION = 3;
    public static final int GRAB_ACTION = 4;
    public static final int SHOOT_ACTION = 5;

    // Is this possible legal?
    private boolean legalPosition(int x, int y) {
        if (x < 0) return false;
        if (y < 0) return false;
        if (x > (mazeSize - 1)) return false;
        if (y > (mazeSize - 1)) return false;
        return true;
    }

    // Does this location have gold?
    private boolean hasGold(int x, int y) {
        return ((maze[x][y] & GOLD_BIT) != 0);
    }

    // Does this location have a pit?
    private boolean hasPit(int x, int y) {
        return ((maze[x][y] & PIT_BIT) != 0);
    }

    // Does this location have a wumpus?
    private boolean hasWumpus(int x, int y) {
        return ((maze[x][y] & WUMPUS_BIT) != 0);
    }

    // Execute a grab of gold at a position
    private boolean grabGold(int x, int y) {
        if (hasGold(x,y)) {
            maze[x][y] -= GOLD_BIT;
            return true;
        }
        else {
            return false;
        }
    }

    // Execute remove wumpus (after successfully shooting it) at this position
    private boolean removeWumpus(int x, int y) {
        if (hasWumpus(x,y)) {
            maze[x][y] -= WUMPUS_BIT;
            return true;
        }
        else {
            return false;
        }
    }

    // Show the actual board (used for debugging purposes)
    private void showBoard() {
        for (int y = mazeSize-1; y >= 0; y--) {
            for (int x = 0; x < mazeSize; x++) {
                System.out.print("|");
                if (hasGold(x,y))
                    System.out.print("G");
                else
                    System.out.print(" ");
                if (hasPit(x,y))
                    System.out.print("P");
                else
                    System.out.print(" ");
                if (hasWumpus(x,y))
                    System.out.print("W");
                else
                    System.out.print(" ");

            }
            System.out.println("|");
        }
    }

    // Routine to set up initial wumpus board
    private void initialize(int size) {
        mazeSize = size;
        hasArrow = true;
        xLocation = 0;
        yLocation = 0;
        facingAngle = 0;
        gameOver = false;
        totalReward = 0;
        maze = new int[mazeSize][mazeSize];
        for (int i = 0; i < mazeSize; i++)
            for (int j = 0; j < mazeSize; j++)
                maze[i][j] = 0;

        // Make sure there is some path from the initial spot to the goal
        maze[0][0] = -1;
        maze[mazeSize-1][mazeSize-1] = -1;
        int x = 0;
        int y = 0;
        while ((x < (mazeSize - 1)) && (y < (mazeSize - 1))) {
            if (Math.random() < 0.5) {
                x = x + 1;
            }
            else {
                y = y + 1;
            }
            maze[x][y] = -1;
        }
        if (x < (mazeSize - 1)) {
            for (int i = x; i < (mazeSize - 1); i++)
                maze[i][mazeSize-1] = -1;
        }
        else if (y < (mazeSize - 1)) {
            for (int i = y; i < (mazeSize - 1); i++)
                maze[mazeSize-1][i] = -1;
        }

        // Place the wumpus randomly
        boolean done = false;
        while (!done) {
            x = (int) Math.floor(Math.random() * mazeSize);
            y = (int) Math.floor(Math.random() * mazeSize);
            if (maze[x][y] != -1) {
                maze[x][y] = maze[x][y] | WUMPUS_BIT;
                done = true;
            }
        }

        // Place mazeSize - 1 pits randomly
        for (int i = 0; i < (mazeSize - 1); i++) {
            done = false;
            while (!done) {
                x = (int) Math.floor(Math.random() * mazeSize);
                y = (int) Math.floor(Math.random() * mazeSize);
                if (maze[x][y] != -1) {
                    if ((maze[x][y] & PIT_BIT) == 0) {
                        maze[x][y] = maze[x][y] | PIT_BIT;
                        done = true;
                    }
                }
            }           
        }

        // Place an item of gold randomly
        done = false;
        while (!done) {
            x = (int) Math.floor(Math.random() * mazeSize);
            y = (int) Math.floor(Math.random() * mazeSize);
            if (maze[x][y] == 0) {
                maze[x][y] = maze[x][y] | GOLD_BIT;
                done = true;
            }
        }
        for (int i = 0; i < mazeSize; i++)
            for (int j = 0; j < mazeSize; j++)
                if (maze[i][j] == -1) { maze[i][j] = 0; }

        showBoard();
    }

    /**
     * Create an initial maze of size 3 if no argument is passed
     */
    public WumpusMaze() {
        initialize(3);
    }

    /**
     * Create an initial maze of size size
     *
     * @param size
     */
    public WumpusMaze(int size) {
        initialize(size);
    }

    /**
     * Execute the action numbered actionNumber
     *
     * @param actionNumber
     */
    public void doAction(int actionNumber) {
        int deltaX;
        int deltaY;
        switch (actionNumber) {
            case FORWARD_ACTION:
                deltaX = (int) Math.cos(Math.toRadians(facingAngle));
                deltaY = (int) Math.sin(Math.toRadians(facingAngle));
                if (legalPosition(xLocation + deltaX,yLocation + deltaY)) {
                    xLocation += deltaX;
                    yLocation += deltaY;
                    if (hasWumpus(xLocation,yLocation)) {
                        gameOver = true;
                        System.out.println("You ran into the wumpus!");
                        totalReward -= LOSE_COST;
                    }
                    else if (hasPit(xLocation,yLocation)) {
                        gameOver = true;
                        System.out.println("You fell into a pit!");
                        totalReward -= LOSE_COST;
                    }
                    else if ((xLocation == (mazeSize - 1)) && (yLocation == (mazeSize - 1))) {
                        System.out.println("You made it to the goal!");
                        gameOver = true;
                    }
                    else {
                        totalReward -= ACTION_COST;
                    }
                }
                break;
            case LEFT_ACTION:
                facingAngle += 90;
                if (facingAngle >= 360) facingAngle -= 360;
                totalReward -= ACTION_COST;
                break;
            case RIGHT_ACTION:
                facingAngle -= 90;
                if (facingAngle < 0) facingAngle += 360;
                totalReward -= ACTION_COST;
                break;
            case GRAB_ACTION:
                if (grabGold(xLocation,yLocation))
                    totalReward += GOLD_REWARD;
                else
                    totalReward -= ACTION_COST;
                break;
            case SHOOT_ACTION:
                if (hasArrow) {
                    deltaX = (int) Math.cos(Math.toRadians(facingAngle));
                    deltaY = (int) Math.sin(Math.toRadians(facingAngle));
                    int x = xLocation + deltaX;
                    int y = yLocation + deltaY;
                    boolean stop = false;
                    while (legalPosition(x,y) && (!stop)) {
                        if (hasWumpus(x,y)) {
                            stop = true;
                            System.out.println("You hit the wumpus!");
                            removeWumpus(x,y);
                        }
                        else {
                            x = x + deltaX;
                            y = y + deltaY;
                        }
                    }
                    hasArrow = false;
                    totalReward -= USE_ARROW_COST;
                }
                else {
                    totalReward -= ACTION_COST;
                }
                break;

        }
    }

    /**
     * Return the current X location
     *
     * @return
     */
    public int currentX() {
        return xLocation;
    }

    /**
     * Return the current Y location
     *
     * @return
     */
    public int currentY() {
        return yLocation;
    }

    /**
     * Return the current facing angle (0 = right, 90 = up, 180 = left, 270 = down)
     *
     * @return
     */
    public int facingAngle() {
        return facingAngle;
    }

    /**
     * Return the value of the gameOver flag
     * @return
     */
    public boolean gameOver() {
        return gameOver;
    }

    /**
     * Return the value of totalReward
     *
     * @return
     */
    public int totalReward() {
        return totalReward;
    }

    /**
     * Return true if the user has an arrow
     *
     * @return
     */
    public boolean hasArrow() {
        return hasArrow;
    }

    /**
     * Return true if the Glitter percept is true at the current location
     *
     * @return
     */
    public boolean hasGlitter() {
        if (hasGold(xLocation,yLocation)) return true;
        return false;
    }

    /**
     * Return true if the Smell percept is true at the current location
     *
     * @return
     */
    public boolean hasSmell() {
        if (legalPosition(xLocation-1,yLocation)) {
            if (hasWumpus(xLocation-1,yLocation)) return true;
        }
        if (legalPosition(xLocation+1,yLocation)) {
            if (hasWumpus(xLocation+1,yLocation)) return true;
        }
        if (legalPosition(xLocation,yLocation-1)) {
            if (hasWumpus(xLocation,yLocation-1)) return true;
        }
        if (legalPosition(xLocation,yLocation+1)) {
            if (hasWumpus(xLocation,yLocation+1)) return true;
        }
        return false;
    }

    /**
     * Return true if the Breeze percept is true at the current location
     *
     * @return
     */
    public boolean hasBreeze() {
        if (legalPosition(xLocation-1,yLocation)) {
            if (hasPit(xLocation-1,yLocation)) return true;
        }
        if (legalPosition(xLocation+1,yLocation)) {
            if (hasPit(xLocation+1,yLocation)) return true;
        }
        if (legalPosition(xLocation,yLocation-1)) {
            if (hasPit(xLocation,yLocation-1)) return true;
        }
        if (legalPosition(xLocation,yLocation+1)) {
            if (hasPit(xLocation,yLocation+1)) return true;
        }
        return false;
    }
}
